<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>Exception Details</title>
    <style>
    {{ css }}
    </style>

    <script>
    {{ javascript }}
    </script>

</head>
<body>
    {% include 'overview.html' %}
    {% include 'exceptions_by_type.html' %}
    {% include 'exceptions_by_model.html' %}
    {% include 'exceptions_by_question_name.html' %}
    {% include 'interviews.html' %}
    {% include 'performance_plot.html' %}
</body>
</html>