{# Answering Instructions #}
Please respond with valid JSON, formatted like so:
{% if include_comment %}
    {"answer": [<put comma-separated list here>], "comment": "<put explanation here>"}
{% else %}
    {"answer": [<put comma-separated list here>]}
{% endif %}

